@@ -19,6 +19,7 @@ class QETxDetails(QObject, QtEventListener):
     _logger = get_logger(__name__)
 
     confirmRemoveLocalTx = pyqtSignal([str], arguments=['message'])
+    txRemoved = pyqtSignal()
     saveTxError = pyqtSignal([str,str], arguments=['code', 'message'])
     saveTxSuccess = pyqtSignal()
 
@@ -55,7 +56,7 @@ class QETxDetails(QObject, QtEventListener):
         self._is_unrelated = False
         self._is_complete = False
         self._is_mined = False
-        self._is_final = False
+        self._is_rbf_enabled = False
         self._lock_delay = 0
         self._sighash_danger = TxSighashDanger()
 
@@ -83,6 +84,12 @@ class QETxDetails(QObject, QtEventListener):
             self._logger.debug(f'new_transaction event for our txid {self._txid}')
             self.update()
 
+    @event_listener
+    def on_event_removed_transaction(self, wallet, tx):
+        if wallet == self._wallet.wallet and tx.txid() == self._txid:
+            self._logger.debug(f'removed my transaction {tx.txid()}')
+            self.txRemoved.emit()
+
     walletChanged = pyqtSignal()
     @pyqtProperty(QEWallet, notify=walletChanged)
     def wallet(self):
@@ -244,8 +251,8 @@ class QETxDetails(QObject, QtEventListener):
         return self._is_complete
 
     @pyqtProperty(bool, notify=detailsChanged)
-    def isFinal(self):
-        return self._is_final
+    def isRbfEnabled(self):
+        return self._is_rbf_enabled
 
     @pyqtProperty(int, notify=detailsChanged)
     def lockDelay(self):
@@ -339,7 +346,7 @@ class QETxDetails(QObject, QtEventListener):
                 self._lnamount.satsInt = 0
 
         self._is_complete = self._tx.is_complete()
-        self._is_final = self._tx.is_final()
+        self._is_rbf_enabled = self._tx.is_rbf_enabled()
         self._is_unrelated = txinfo.amount is None and self._lnamount.isEmpty
         self._is_lightning_funding_tx = txinfo.is_lightning_funding_tx
         self._can_broadcast = txinfo.can_broadcast
@@ -426,7 +433,7 @@ class QETxDetails(QObject, QtEventListener):
         self._can_broadcast = False
         self.detailsChanged.emit()
 
-    @pyqtSlot(str,str,str)
+    @pyqtSlot(str, str, str)
     def onBroadcastFailed(self, txid, code, reason):
         if txid != self._txid:
             return
@@ -461,7 +468,6 @@ class QETxDetails(QObject, QtEventListener):
 
         # NOTE: from here, the tx/txid is unknown and all properties are invalid.
         # UI should close TxDetails and avoid interacting with this qetxdetails instance.
-        self._txid = None
         self._tx = None
 
     @pyqtSlot()
